-- If you change this file, run: docker container rm local_ubs_postgres_data_1 local_ubs_postgres_1

-- Specify that everything is happening in the UBS DB. Must be in every SQL init script!
\c ubs;

-- IMPORTANT! User is different in CI.
-- Here we assume the permissions of the user executing the command.
-- The user running this command is the same one that will be shared by
-- all production apps.
CREATE SCHEMA IF NOT EXISTS ubs_state;
SET search_path TO ubs_state; -- Make everything in the new schema

-- magic table, equivalent to the magic table in UBS
-- this table is activated when UBAP is decoupled from UBS
CREATE TABLE IF NOT EXISTS magic (
  id SERIAL PRIMARY KEY NOT NULL,
  value varchar(1024) NOT NULL,
  sha256_value bytea NOT NULL CHECK(length(sha256_value) = 32) UNIQUE,
  date_created TIMESTAMPTZ NOT NULL DEFAULT NOW()
);

CREATE INDEX IF NOT EXISTS idx_magic_sha256_value ON magic (sha256_value);

-- Entry point and link to the world
-- Using sha256, not a FK to file_index or file_details, b/c it's possible to only have the sha256, e.g. APC lookup
-- It's assumed that these are the sha256s of the full files. If we run on truncated file, we'll need to rerun
-- everything under main.
CREATE TABLE IF NOT EXISTS main (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  sha256 BYTEA NOT NULL CHECK(length(sha256) = 32),
  sha256_truncated BYTEA,
  s3_key varchar(1024),
  truncated BOOLEAN,
  magic_id INT REFERENCES magic(id),
  date_created TIMESTAMPTZ NOT NULL DEFAULT NOW(),
  date_modified TIMESTAMPTZ
);
CREATE UNIQUE INDEX IF NOT EXISTS idx_main_sha256 ON main (sha256);
CREATE INDEX IF NOT EXISTS idx_file_main_date_created ON main (date_created);

-- Add new columns to main if they don't exist yet
ALTER TABLE main
    ADD COLUMN IF NOT EXISTS sha256_truncated BYTEA,
    ADD COLUMN IF NOT EXISTS s3_key CHARACTER VARYING(1024),
    ADD COLUMN IF NOT EXISTS truncated BOOLEAN,
    ADD COLUMN IF NOT EXISTS magic_id INT REFERENCES magic(id),
    ADD COLUMN IF NOT EXISTS date_modified timestamptz;

-- Create partner table
CREATE TABLE IF NOT EXISTS partner (
  id SMALLSERIAL PRIMARY KEY NOT NULL,
  name VARCHAR(128) NOT NULL,
  date_created TIMESTAMPTZ NOT NULL DEFAULT NOW()
);

INSERT INTO partner (id, name) VALUES (0, 'apc') ON CONFLICT DO NOTHING;

CREATE UNIQUE INDEX IF NOT EXISTS idx_partner_name ON partner(name);

-- Create main partner link table
CREATE TABLE IF NOT EXISTS main_partner_link (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  partner_id SMALLINT NOT NULL REFERENCES partner(id),
  main_id BIGINT NOT NULL REFERENCES main(id),
  response_state SMALLINT NOT NULL,
  date_created TIMESTAMPTZ NOT NULL DEFAULT NOW(),
  date_modified TIMESTAMPTZ NOT NULL DEFAULT NOW()
);

CREATE INDEX IF NOT EXISTS idx_main_partner_link_partner_id ON main_partner_link(partner_id);
CREATE INDEX IF NOT EXISTS idx_main_partner_link_main_id ON main_partner_link(main_id);
CREATE UNIQUE INDEX IF NOT EXISTS idx_main_id_partner_id ON main_partner_link(partner_id, main_id);

CREATE TABLE IF NOT EXISTS version (
  id SMALLSERIAL PRIMARY KEY NOT NULL,
  major SMALLINT NOT NULL,
  minor SMALLINT NOT NULL,
  date_created TIMESTAMPTZ NOT NULL DEFAULT NOW()
);
CREATE UNIQUE INDEX IF NOT EXISTS idx_version_major_minor ON version (major, minor);

CREATE TABLE IF NOT EXISTS status (
  id SMALLSERIAL PRIMARY KEY NOT NULL,
  name VARCHAR NOT NULL,
  date_created TIMESTAMPTZ NOT NULL DEFAULT NOW()
);
CREATE UNIQUE INDEX IF NOT EXISTS idx_status_name ON status(name);

-- Details of a specific stage
CREATE TABLE IF NOT EXISTS stage (
  id SMALLSERIAL PRIMARY KEY NOT NULL,
  name VARCHAR NOT NULL,
  current_version_id SMALLINT NOT NULL REFERENCES version (id),
  requires_file BOOLEAN NOT NULL DEFAULT FALSE,
  date_created TIMESTAMPTZ NOT NULL DEFAULT NOW(),
  date_modified TIMESTAMPTZ NOT NULL DEFAULT NOW()
);
CREATE UNIQUE INDEX IF NOT EXISTS idx_stage_name ON stage (name);

-- Relates main to a stage
CREATE TABLE IF NOT EXISTS main_stage_link (
    id BIGSERIAL PRIMARY KEY NOT NULL,
    main_id BIGINT NOT NULL REFERENCES main(id),
    stage_id SMALLINT NOT NULL REFERENCES stage(id),
    date_created TIMESTAMPTZ NOT NULL DEFAULT NOW()
);
CREATE UNIQUE INDEX IF NOT EXISTS idx_main_stage_link_main_id_stage_id ON main_stage_link (main_id, stage_id);

-- Only stores details of current state per version
-- Possible to have multiple executions of the same stage, but only current state is known
CREATE TABLE IF NOT EXISTS stage_status (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  main_stage_link_id BIGINT NOT NULL REFERENCES main_stage_link(id),
  version_id SMALLINT NOT NULL REFERENCES version(id),
  status_id SMALLINT NOT NULL REFERENCES status(id),
  completion_date TIMESTAMPTZ NOT NULL DEFAULT NOW(), -- Date reported by the stage
  date_created TIMESTAMPTZ NOT NULL DEFAULT NOW(),
  date_modified TIMESTAMPTZ NOT NULL DEFAULT NOW()
);
CREATE UNIQUE INDEX IF NOT EXISTS idx_stage_status_main_stage_link_id_version ON stage_status (main_stage_link_id, version_id);

-- Target stages are the stages that are unblocked by this stage
CREATE TABLE IF NOT EXISTS stage_target (
    id SMALLSERIAL PRIMARY KEY NOT NULL,
    stage_id SMALLINT NOT NULL REFERENCES stage(id), -- The current stage
    target_stage_id SMALLINT NOT NULL REFERENCES stage(id), -- The stages that are unblocked by stage_id's completion
    date_created TIMESTAMPTZ NOT NULL DEFAULT NOW(),
    date_modified TIMESTAMPTZ NOT NULL DEFAULT NOW()
);
CREATE UNIQUE INDEX IF NOT EXISTS idx_stage_target_stage_id_target_stage_id ON stage_target (stage_id, target_stage_id);

-- The topic that triggers the specified stage
-- 1:1 relationship between topic and stage
-- Note: not unique on sns_arn so we can have duplicate stages to avoid circles,
-- e.g. apc lookup -> apc upload -> apc lookup
CREATE TABLE IF NOT EXISTS stage_topic (
    id SMALLSERIAL PRIMARY KEY NOT NULL,
    stage_id SMALLINT NOT NULL REFERENCES stage(id),
    sns_arn VARCHAR NOT NULL,
    enabled BOOLEAN NOT NULL DEFAULT FALSE,
    date_created TIMESTAMPTZ NOT NULL DEFAULT NOW(),
    date_modified TIMESTAMPTZ NOT NULL DEFAULT NOW()
);
CREATE UNIQUE INDEX IF NOT EXISTS idx_stage_topic_stage_id ON stage_topic (stage_id);

-- Details of a specific pipeline
CREATE TABLE IF NOT EXISTS pipeline (
  id SMALLSERIAL PRIMARY KEY NOT NULL,
  name VARCHAR NOT NULL,
  current_version_id SMALLINT NOT NULL REFERENCES version(id),
  date_created TIMESTAMPTZ NOT NULL DEFAULT NOW(),
  date_modified TIMESTAMPTZ NOT NULL DEFAULT NOW()
);
CREATE UNIQUE INDEX IF NOT EXISTS idx_pipeline_name ON pipeline (name);

-- Relates a hash to a pipeline
CREATE TABLE IF NOT EXISTS main_pipeline_link (
    id BIGSERIAL PRIMARY KEY NOT NULL,
    main_id BIGINT NOT NULL REFERENCES main(id),
    pipeline_id SMALLINT NOT NULL REFERENCES pipeline(id),
    date_created TIMESTAMPTZ NOT NULL DEFAULT NOW()
);
CREATE UNIQUE INDEX IF NOT EXISTS idx_main_pipeline_link_main_id_pipeline_id ON main_pipeline_link (main_id, pipeline_id);

-- Shows the specific status of a pipeline per hash
-- 1:many sha256:pipeline_status, but 1:1 per version
CREATE TABLE IF NOT EXISTS pipeline_status (
    id BIGSERIAL PRIMARY KEY NOT NULL,
    main_pipeline_link_id BIGINT NOT NULL REFERENCES main_pipeline_link(id),
    version_id SMALLINT NOT NULL REFERENCES version(id),
    status_id SMALLINT NOT NULL REFERENCES status(id),
    completion_date TIMESTAMPTZ NOT NULL DEFAULT NOW(), -- Date reported by the stage that caused the pipeline to finish
    date_created TIMESTAMPTZ NOT NULL DEFAULT NOW(),
    date_modified TIMESTAMPTZ NOT NULL DEFAULT NOW()
);
CREATE UNIQUE INDEX IF NOT EXISTS idx_pipeline_status_main_pipeline_link_id_version_id ON pipeline_status (main_pipeline_link_id, version_id);

-- Adds stages to a pipeline
CREATE TABLE IF NOT EXISTS pipeline_stage_link (
    id SMALLSERIAL PRIMARY KEY NOT NULL,
    pipeline_id SMALLINT NOT NULL REFERENCES pipeline(id),
    stage_id SMALLINT NOT NULL REFERENCES stage(id),
    date_created TIMESTAMPTZ NOT NULL DEFAULT NOW()
);
CREATE UNIQUE INDEX IF NOT EXISTS idx_pipeline_stage_link_pipeline_id_stage_id ON pipeline_stage_link (pipeline_id, stage_id);

--- get magic and file details of the hash --
CREATE OR REPLACE FUNCTION get_magic_and_file_details(in_sha256 bytea)
    RETURNS TABLE(sha256_truncated bytea, s3_key varchar, truncated boolean, magic_value varchar)
AS $$
    BEGIN
        RETURN QUERY

        SELECT
            m.sha256_truncated,
            m.s3_key,
            m.truncated,
            mgc.value
        FROM ubs_state.main m
            JOIN ubs_state.magic mgc
            ON m.sha256=in_sha256 AND m.magic_id=mgc.id;

    END
$$ LANGUAGE plpgsql;

---- update the magic and file details for the hash --
CREATE OR REPLACE FUNCTION update_magic_and_file_details(
                                                        in_sha256 bytea,
                                                        in_sha256_truncated bytea,
                                                        in_s3_key varchar,
                                                        in_truncated boolean,
                                                        in_magic_value varchar,
                                                        in_magic_sha256 bytea)
RETURNS void
AS $$
    DECLARE
        in_magic_id INT;
    BEGIN
        -- if magic value exist, get the magic id; otherwise insert new row and get the magic id
        INSERT INTO ubs_state.magic(value, sha256_value) VALUES (in_magic_value, in_magic_sha256)
            ON CONFLICT (sha256_value) DO NOTHING ;

        SELECT id from ubs_state.magic WHERE sha256_value = in_magic_sha256 INTO in_magic_id;

        -- if sha256 exists in main, update the row, otherwise insert a new row
        INSERT INTO ubs_state.main(sha256, sha256_truncated, s3_key, truncated, magic_id)
            VALUES (in_sha256, in_sha256_truncated, in_s3_key, in_truncated, in_magic_id)
            ON CONFLICT (sha256) DO UPDATE
                SET sha256_truncated = in_sha256_truncated,
                    s3_key = in_s3_key,
                    truncated = in_truncated,
                    magic_id = in_magic_id,
                    date_modified = now();

    END
$$ LANGUAGE plpgsql;

---- creating statuses, initial pipelines, stage definitions, etc. ----
DO $$
DECLARE
    version_1_id smallint;
    version_2_id smallint;

    bsv_stage_id smallint;
    apc_lookup_stage_id smallint;
    apc_lookup_poll_stage_id smallint;
    apc_upload_stage_id smallint;
    yara_stage_id smallint;
    file_exists_stage_id smallint;
    apc_upload_allowed_stage_id smallint;

    reputation_pipeline_id smallint;
    tau_pipeline_id smallint;
    reputation_no_file_share_id smallint;

BEGIN
    INSERT INTO status (name) VALUES
        ('NOT_STARTED'),
        ('BLOCKED'),
        ('IN_PROGRESS'),
        ('DECLINE_EXECUTE'),
        ('COMPLETE'),
        ('ERROR_RETRY'),
        ('ERROR')
    ON CONFLICT DO NOTHING;

    INSERT INTO version (major, minor) VALUES
        (1, 0),
        (2, 0)
    ON CONFLICT DO NOTHING;

    version_1_id := (SELECT id FROM ubs_state.version WHERE major = 1 AND minor = 0);
    version_2_id := (SELECT id FROM ubs_state.version WHERE major = 2 AND minor = 0);

    INSERT INTO stage (name, current_version_id, requires_file) VALUES
        ('BINARY_STORAGE_VALIDATION', version_1_id, FALSE),
        ('APC_LOOKUP', version_1_id, FALSE),
        ('APC_LOOKUP_POLL', version_1_id, FALSE),
        ('APC_UPLOAD', version_1_id, TRUE),
        ('YARA', version_1_id, TRUE),
        ('FILE_EXISTS', version_1_id, TRUE),
        ('APC_UPLOAD_ALLOWED', version_1_id, TRUE)
    ON CONFLICT DO NOTHING;

    bsv_stage_id := (SELECT id FROM ubs_state.stage WHERE name = 'BINARY_STORAGE_VALIDATION');
    apc_lookup_stage_id := (SELECT id FROM ubs_state.stage WHERE name = 'APC_LOOKUP');
    apc_lookup_poll_stage_id := (SELECT id FROM ubs_state.stage WHERE name = 'APC_LOOKUP_POLL');
    apc_upload_stage_id := (SELECT id FROM ubs_state.stage WHERE name = 'APC_UPLOAD');
    yara_stage_id := (SELECT id FROM ubs_state.stage WHERE name = 'YARA');
    file_exists_stage_id := (SELECT id FROM ubs_state.stage WHERE name = 'FILE_EXISTS');
    apc_upload_allowed_stage_id := (SELECT id FROM ubs_state.stage WHERE name = 'APC_UPLOAD_ALLOWED');

    INSERT INTO stage_target (stage_id, target_stage_id) VALUES
        (file_exists_stage_id, apc_upload_stage_id),
        (apc_upload_allowed_stage_id, apc_upload_stage_id),
        (apc_lookup_stage_id, apc_upload_stage_id),
        (apc_upload_stage_id, apc_lookup_poll_stage_id),
        (file_exists_stage_id, yara_stage_id)
    ON CONFLICT DO NOTHING;

    INSERT INTO pipeline (name, current_version_id) VALUES
        ('REPUTATION', version_2_id),
        ('TAU', version_1_id),
        ('REPUTATION_NO_FILE_SHARE', version_2_id)
    ON CONFLICT DO NOTHING;

    reputation_pipeline_id := (SELECT id FROM ubs_state.pipeline WHERE name = 'REPUTATION');
    tau_pipeline_id := (SELECT id FROM ubs_state.pipeline WHERE name = 'TAU');
    reputation_no_file_share_id := (SELECT id FROM ubs_state.pipeline WHERE name = 'REPUTATION_NO_FILE_SHARE');

    INSERT INTO pipeline_stage_link (pipeline_id, stage_id) VALUES
        (reputation_pipeline_id, apc_lookup_stage_id),
        (reputation_pipeline_id, apc_upload_stage_id),
        (reputation_pipeline_id, apc_lookup_poll_stage_id),
        (reputation_pipeline_id, file_exists_stage_id),
        (reputation_pipeline_id, apc_upload_allowed_stage_id),
        (tau_pipeline_id, file_exists_stage_id),
        (tau_pipeline_id, yara_stage_id),
        (reputation_no_file_share_id, file_exists_stage_id),
        (reputation_no_file_share_id, apc_lookup_stage_id)
    ON CONFLICT DO NOTHING;
END $$;

-- get pipeline status --
CREATE OR REPLACE FUNCTION get_pipeline_status(in_pipeline_id smallint,
                                               in_main_id bigint)
RETURNS varchar AS
$$
    DECLARE
        var_stage_rec RECORD;
        var_stage_status_name varchar;
        var_state smallint;
        var_final_pipeline_state varchar;
        var_has_complete_state bool;

    BEGIN
        var_state := 7;
        var_has_complete_state := false;

        FOR var_stage_rec IN (
            SELECT stage_id
            FROM ubs_state.pipeline_stage_link
            WHERE pipeline_id = in_pipeline_id
        )
        LOOP
            WITH get_main_stage_id AS (
                SELECT id
                FROM ubs_state.main_stage_link
                WHERE main_id = in_main_id AND stage_id = var_stage_rec.stage_id
            ), get_status_id AS (
                SELECT ss.status_id
                FROM ubs_state.stage s
                JOIN ubs_state.stage_status ss ON s.current_version_id = ss.version_id
                JOIN get_main_stage_id gmsi ON ss.main_stage_link_id = gmsi.id
                WHERE s.id = var_stage_rec.stage_id
            )
            SELECT name INTO var_stage_status_name
            FROM ubs_state.status s
            JOIN get_status_id gsi
            ON s.id = gsi.status_id;

            -- priority: ERROR -> BLOCKED -> ERROR_RETRY -> IN_PROGRESS -> NOT_STARTED OR COMPLETE/DECLINE_EXECUTE
            CASE var_stage_status_name
            WHEN 'ERROR' THEN
                var_state = 1;
                var_final_pipeline_state = 'ERROR';
            WHEN 'BLOCKED' THEN
                IF var_state > 2 THEN
                    var_state = 2;
                    var_final_pipeline_state = 'BLOCKED';
                END IF;
            WHEN 'ERROR_RETRY' THEN
                IF var_state > 3 THEN
                    var_state = 3;
                    var_final_pipeline_state = 'BLOCKED';
                END IF;
            WHEN 'IN_PROGRESS' THEN
                IF var_state > 4 THEN
                    var_state = 4;
                    var_final_pipeline_state = 'IN_PROGRESS';
                END IF;
            WHEN 'NOT_STARTED' THEN
                IF var_state > 5 THEN
                    var_state = 5;
                    var_final_pipeline_state = 'NOT_STARTED';
                END IF;
            WHEN NULL THEN
                RAISE INFO 'stage_id (%) in pipeline id (%) does not have status, consider it as NOT_STARTED',
                    var_stage_rec.stage_id, in_pipeline_id;
                if var_state > 5 then
                    var_state = 5;
                    var_final_pipeline_state = 'NOT_STARTED';
                end if;
            WHEN 'DECLINE_EXECUTE' THEN
                var_has_complete_state = true;
                IF var_state > 6 THEN
                    var_state = 6;
                    var_final_pipeline_state = 'COMPLETE';
                END IF;
            WHEN 'COMPLETE' THEN
                var_has_complete_state = true;
                IF var_state > 6 THEN
                    var_state = 6;
                    var_final_pipeline_state = 'COMPLETE';
                END IF;
            ELSE
                RAISE WARNING 'stage_id (%) in pipeline id (%) has an undefined status (%), consider it as NOT_STARTED',
                    var_stage_rec.stage_id, in_pipeline_id, var_stage_status_name;
                 if var_state > 5 then
                    var_state = 5;
                    var_final_pipeline_state = 'NOT_STARTED';
                end if;
            END CASE;
        END LOOP;

        IF var_final_pipeline_state = 'NOT_STARTED' AND var_has_complete_state THEN
            var_final_pipeline_state = 'IN_PROGRESS';
        END IF;
        RETURN var_final_pipeline_state;
    END
$$ LANGUAGE plpgsql;

-- stage has a version update, now update the downstream stages statuses to NOT_STARTED
CREATE OR REPLACE FUNCTION update_pipeline_ver(in_pipeline_id smallint)
    RETURNS void AS
$$
DECLARE
    var_pl_cur_ver RECORD;

BEGIN
    SELECT v.major, v.minor INTO var_pl_cur_ver
    FROM ubs_state.pipeline p
             JOIN ubs_state.version v on p.id = in_pipeline_id AND p.current_version_id = v.id;

    -- insert new version number if not exists; update the pipeline version
    INSERT INTO ubs_state.version(major, minor,date_created)
    VALUES (var_pl_cur_ver.major, var_pl_cur_ver.minor + 1, NOW())
    ON CONFLICT (major, minor) DO NOTHING ;

    WITH get_pl_version AS (
        SELECT id
        FROM ubs_state.version
        WHERE major = var_pl_cur_ver.major AND minor = var_pl_cur_ver.minor + 1
    )
    UPDATE ubs_state.pipeline p
    SET current_version_id = gpv.id
    FROM get_pl_version gpv
    WHERE p.id = in_pipeline_id;

END
$$ LANGUAGE plpgsql;


 -- stage has a version update, now update the downstream stages statuses to NOT_STARTED
CREATE OR REPLACE FUNCTION update_ds_stage_status(in_cur_stage_id smallint,
                                                      in_main_id bigint)
RETURNS void AS
$$
    DECLARE
        var_not_started_status_id smallint;
        var_ds_stage RECORD;
        var_main_stage_link_id bigint;
        var_version_id smallint;

    BEGIN
        -- get all downstream stages --
        CREATE TEMP TABLE IF NOT EXISTS ds_stages(id bigint primary key);
        TRUNCATE TABLE ds_stages;

        SELECT id INTO var_not_started_status_id
        FROM ubs_state.status WHERE name = 'NOT_STARTED';

        WITH RECURSIVE get_downstream_ids_rec(id) AS (
            SELECT target_stage_id
            FROM ubs_state.stage_target
            WHERE stage_id = in_cur_stage_id

            UNION

            SELECT st.target_stage_id
            FROM ubs_state.stage_target st
            INNER JOIN get_downstream_ids_rec gdi
            ON st.stage_id = gdi.id
        )
        INSERT INTO ds_stages
            SELECT id FROM get_downstream_ids_rec;

        -- update all the downstream stages to NOT_STARTED --
        FOR var_ds_stage IN (
            SELECT DISTINCT *
            FROM ds_stages
        )
        LOOP
            INSERT INTO ubs_state.main_stage_link (main_id, stage_id, date_created)
            VALUES (in_main_id, var_ds_stage.id, now())
            ON CONFLICT (main_id, stage_id) DO NOTHING ;

            select id into var_main_stage_link_id
            from ubs_state.main_stage_link
                   where main_id = in_main_id and stage_id = var_ds_stage.id;

            select current_version_id into var_version_id
            from ubs_state.stage
                      where id = var_ds_stage.id;

            INSERT INTO ubs_state.stage_status(main_stage_link_id, version_id, status_id, date_created, date_modified)
            VALUES (var_main_stage_link_id, var_version_id, var_not_started_status_id, now(), now())
            ON CONFLICT (main_stage_link_id, version_id)
                DO UPDATE
                    SET status_id = var_not_started_status_id,
                        date_modified = NOW();
        END LOOP;

    END
$$ LANGUAGE plpgsql;


-- update stage and related pipelines statuses per stage input and conditions
CREATE OR REPLACE FUNCTION update_stage_pipeline_status(in_sha256 bytea,
                                                        in_stage_name varchar,
                                                        in_stage_status varchar,
                                                        in_stage_version_major smallint,
                                                        in_stage_version_minor smallint,
                                                        in_stage_completion_date timestamp with time zone)
RETURNS bool AS
$$
    DECLARE
        var_stage_cur_stage RECORD;
        var_cur_stage_completion_date timestamp with time zone;
        var_stage_version_updated BOOL;
        var_stage_new_ver_id smallint;
        var_stage_status_id smallint;
        var_main_id bigint;
        var_main_stage_link_id bigint;
        var_pipeline_rec RECORD;
        var_new_pipeline_status varchar;
        var_pipeline_status_id bigint;

    BEGIN
        var_stage_version_updated := false;

        IF NOT EXISTS (SELECT 1
                       FROM ubs_state.stage
                       WHERE name = in_stage_name) THEN
            RAISE WARNING 'stage name does not exist';
            RETURN false;
        END IF;

        IF NOT EXISTS (SELECT 1
                       FROM ubs_state.status
                       WHERE name = in_stage_status) THEN
            RAISE WARNING 'status name does not exist';
            RETURN false;
        end if;

        -- Insert only if not exists to reduce dead tuples introduced by CONFLICT DP NOTHING statement
        SELECT id INTO var_main_id
        FROM ubs_state.main
        WHERE sha256 = in_sha256;

        IF var_main_id IS NULL THEN
            INSERT INTO ubs_state.main(sha256)
            VALUES (in_sha256)
            ON CONFLICT (sha256) DO NOTHING ;

            SELECT id INTO var_main_id
            FROM ubs_state.main
            WHERE sha256 = in_sha256;
        END IF;

        SELECT v.major, v.minor, s.id, s.name INTO var_stage_cur_stage
        FROM ubs_state.stage s
        JOIN ubs_state.version v ON s.name = in_stage_name and s.current_version_id = v.id;

        -- update stage table: update version if input version is newer than the current version for the stage name
        IF var_stage_cur_stage.major < in_stage_version_major OR
           (var_stage_cur_stage.major = in_stage_version_major AND var_stage_cur_stage.minor < in_stage_version_minor) THEN

            var_stage_version_updated = TRUE; -- newer version

            -- Insert only if not exists to reduce dead tuples introduced by CONFLICT DP NOTHING statement
            IF NOT EXISTS(SELECT 1 FROM ubs_state.version WHERE major=in_stage_version_major AND minor=in_stage_version_minor) THEN
                INSERT INTO ubs_state.version(major, minor, date_created)
                VALUES (in_stage_version_major, in_stage_version_minor, NOW())
                ON CONFLICT (major, minor) DO NOTHING;
            END IF;

            WITH get_ver AS (
                SELECT id FROM ubs_state.version
                WHERE major = in_stage_version_major AND minor = in_stage_version_minor
            )
            UPDATE ubs_state.stage s
            SET current_version_id = v.id
            FROM get_ver v WHERE s.name = in_stage_name;

        ELSIF (var_stage_cur_stage.major > in_stage_version_major OR
               (var_stage_cur_stage.major = in_stage_version_major AND var_stage_cur_stage.minor > in_stage_version_minor) ) THEN
            RAISE WARNING 'stage version is older than current version, no stage is updated';
            RETURN false;
        ELSIF (var_stage_cur_stage.major = in_stage_version_major AND var_stage_cur_stage.minor = in_stage_version_minor) THEN
            SELECT ss0.completion_date into var_cur_stage_completion_date
            FROM ubs_state.stage_status ss0
            WHERE EXISTS (
                        SELECT 1
                        FROM ubs_state.stage_status ss
                        JOIN ubs_state.main_stage_link msl ON ss.main_stage_link_id = msl.id
                        JOIN ubs_state.version v ON ss.version_id = v.id
                        WHERE msl.stage_id = var_stage_cur_stage.id AND msl.main_id = var_main_id AND
                              v.major = in_stage_version_major AND v.minor = in_stage_version_minor AND
                              ss0.id = ss.id);
            IF var_cur_stage_completion_date > in_stage_completion_date THEN
                RAISE WARNING 'stage (%) current completion date (%) is newer than requested completion date (%). No update made to stage/pipelines!',
                            in_stage_name, var_cur_stage_completion_date, in_stage_completion_date;
                RETURN false;
            END IF;
        END IF;

        -- update stage_status ---
        SELECT current_version_id INTO var_stage_new_ver_id -- this is the new id if updated
            FROM ubs_state.stage
            WHERE name = in_stage_name;

        SELECT id INTO var_stage_status_id
            FROM ubs_state.status
            WHERE name = in_stage_status;

        SELECT id INTO var_main_stage_link_id
        FROM ubs_state.main_stage_link
        WHERE stage_id=var_stage_cur_stage.id and main_id=var_main_id;

        -- Insert only if not exists to reduce dead tuples introduced by CONFLICT DP NOTHING statement
        IF var_main_stage_link_id IS NULL THEN
            INSERT INTO ubs_state.main_stage_link(main_id, stage_id, date_created)
            VALUES (var_main_id, var_stage_cur_stage.id, NOW())
            ON CONFLICT(main_id, stage_id) DO NOTHING ;

            SELECT id INTO var_main_stage_link_id
            FROM ubs_state.main_stage_link
            WHERE stage_id=var_stage_cur_stage.id and main_id=var_main_id;
        END IF;

        -- Redundant update statement for reducing null tuples on the table
        IF EXISTS(SELECT 1 FROM ubs_state.stage_status
            WHERE main_stage_link_id = var_main_stage_link_id AND version_id = var_stage_new_ver_id) THEN
            UPDATE ubs_state.stage_status SET
              status_id = var_stage_status_id,
              completion_date = in_stage_completion_date,
              date_modified = NOW()
            WHERE main_stage_link_id = var_main_stage_link_id
            AND version_id = var_stage_new_ver_id;
        ELSE
            INSERT INTO ubs_state.stage_status(main_stage_link_id, version_id, status_id, completion_date)
            VALUES (var_main_stage_link_id, var_stage_new_ver_id, var_stage_status_id, in_stage_completion_date)
            ON CONFLICT (main_stage_link_id, version_id)
            DO UPDATE SET
                          status_id = var_stage_status_id,
                          completion_date = in_stage_completion_date,
                          date_modified = NOW();
        END if;

        -- update pipeline status ---
        FOR var_pipeline_rec IN (
            SELECT p.id, p.name, p.current_version_id
            FROM ubs_state.pipeline_stage_link psl
            JOIN ubs_state.pipeline p
            ON psl.stage_id = var_stage_cur_stage.id AND psl.pipeline_id = p.id)
        LOOP
            IF NOT EXISTS (
                SELECT 1
                FROM ubs_state.main_pipeline_link
                WHERE main_id = var_main_id AND pipeline_id = var_pipeline_rec.id
            ) THEN
                RAISE WARNING 'pipeline (%) and hash (%) does not have an entry in main_pipeline_link table. Pipeline NOT updated!',
                     var_pipeline_rec.name, in_sha256;
                CONTINUE;
            END IF;

            if var_stage_version_updated THEN
                -- increase the pipeline version and update downstream stages to NOT_STARTED
                PERFORM ubs_state.update_pipeline_ver(var_pipeline_rec.id);
                PERFORM ubs_state.update_ds_stage_status(var_stage_cur_stage.id,
                                                   var_main_id);
            ELSE -- same as current version
                IF NOT EXISTS (
                        SELECT 1
                        FROM ubs_state.pipeline_status ps
                        JOIN ubs_state.main_pipeline_link mpl ON ps.main_pipeline_link_id = mpl.id
                        WHERE mpl.pipeline_id = var_pipeline_rec.id AND mpl.main_id = var_main_id AND
                              ps.version_id = var_pipeline_rec.current_version_id) THEN
                  RAISE WARNING 'pipeline (%) does not have a current status, nothing to update!', var_pipeline_rec.name;
                  CONTINUE;
                END IF;
            end if;

            -- generate pipeline status --
            IF in_stage_status = 'ERROR_RETRY' OR
               in_stage_status = 'ERROR' OR
               in_stage_status = 'BLOCKED' OR
               in_stage_status = 'IN_PROGRESS' THEN
               var_new_pipeline_status = in_stage_status;
            ELSE
                CASE
                WHEN in_stage_status = 'COMPLETE' OR in_stage_status = 'NOT_STARTED' THEN
                    SELECT ubs_state.get_pipeline_status(var_pipeline_rec.id, var_main_id)
                    into var_new_pipeline_status;
                WHEN in_stage_status = 'DECLINE_EXECUTE' THEN
                    var_new_pipeline_status = 'COMPLETE'; -- early exit
                ELSE
                    RAISE WARNING 'invalid stage status (%)', in_stage_status;
                    RETURN false;
                END CASE;
            END IF;

            -- update pipeline status ---
            SELECT id INTO var_pipeline_status_id
                FROM ubs_state.status
                WHERE name = var_new_pipeline_status;

            WITH get_mpl AS (
                SELECT id
                FROM ubs_state.main_pipeline_link
                WHERE main_id = var_main_id AND pipeline_id = var_pipeline_rec.id
            ), get_pipeline_ver(id) AS (
                SELECT current_version_id
                FROM ubs_state.pipeline
                where id = var_pipeline_rec.id
            )  -- insert newer version pipeline status or update existing pipeline status
            INSERT INTO ubs_state.pipeline_status(main_pipeline_link_id, version_id, status_id, date_created)
            SELECT gmpl.id, gpv.id, var_pipeline_status_id, NOW()
            FROM get_mpl gmpl, get_pipeline_ver gpv
            ON CONFLICT (main_pipeline_link_id, version_id)
                DO UPDATE
                SET status_id = var_pipeline_status_id, date_modified = NOW();
        END LOOP;

    RETURN true;
    END
$$ LANGUAGE plpgsql;


-- check if all dependencies of a given stage have all completed --
CREATE OR REPLACE FUNCTION check_stage_dependencies(in_main_id bigint,
                                                    in_target_id smallint)
RETURNS BOOLEAN AS
$$
    DECLARE
        var_dependent RECORD;
        var_dep_status_name varchar;
        var_is_complete BOOLEAN;
    BEGIN
        var_is_complete := TRUE;

        FOR var_dependent IN (
            SELECT stage_id
            FROM ubs_state.stage_target
            WHERE target_stage_id = in_target_id
        )
        LOOP
            WITH get_msl AS (
                SELECT id
                FROM ubs_state.main_stage_link
                WHERE main_id = in_main_id AND
                      stage_id = var_dependent.stage_id
            ), get_ver(id) AS (
                SELECT current_version_id
                FROM ubs_state.stage
                WHERE id = var_dependent.stage_id
            )
            SELECT sts.name INTO var_dep_status_name
            FROM ubs_state.status sts, get_msl gm, get_ver gv, ubs_state.stage_status ss
            WHERE ss.version_id = gv.id AND ss.status_id = sts.id AND ss.main_stage_link_id = gm.id;

            IF var_dep_status_name IS NULL OR var_dep_status_name <> 'COMPLETE' THEN
                var_is_complete = FALSE;
                EXIT;
            END IF;
        END LOOP;

        RETURN var_is_complete;

    END
$$ LANGUAGE plpgsql;

DO $$ BEGIN
    -- return type of get_next_stage ---
    CREATE TYPE type_next_stage AS (
        target_name varchar,
        target_topic varchar,
        requires_file boolean
    );
EXCEPTION
    WHEN duplicate_object THEN null;
END $$;

-- get next unblocked stages
CREATE OR REPLACE FUNCTION get_next_stages(in_sha256 bytea,
                                           in_stage_name varchar,
                                           in_stage_status varchar,
                                           in_version_major smallint,
                                           in_version_minor smallint)
RETURNS ubs_state.type_next_stage[]
AS $$
    DECLARE
        var_next_stages ubs_state.type_next_stage[];
        var_next_stg ubs_state.type_next_stage;
        var_target RECORD;
        var_cur_stage_id smallint;
        var_main_id bigint;
        var_is_complete BOOLEAN;
        var_target_name_sns_arn_requires_file RECORD;

    BEGIN
        -- check if current version number is correct
        IF NOT EXISTS(
                SELECT v.major, v.minor, v.id
                FROM ubs_state.version v
                JOIN ubs_state.stage s
                ON s.name = in_stage_name AND
                   s.current_version_id = v.id AND
                   v.major = in_version_major AND
                   v.minor = in_version_minor
        ) THEN
            RAISE INFO 'version number (%).(%) is not current for stage (%)',
                in_version_major, in_version_minor, in_stage_name;
            RETURN var_next_stages;
        END IF;

        SELECT id INTO var_main_id
        FROM ubs_state.main WHERE sha256 = in_sha256;

        if (in_stage_status = 'COMPLETE') THEN
            FOR var_target IN (
                SELECT target_stage_id
                FROM ubs_state.stage_target st
                JOIN ubs_state.stage s ON s.name = in_stage_name AND s.id = st.stage_id
            )
            LOOP
                -- check all immediate dependents
                SELECT ubs_state.check_stage_dependencies(var_main_id,
                                                          var_target.target_stage_id)
                    INTO var_is_complete;

                IF var_is_complete THEN
                    SELECT s.name, st.sns_arn, s.requires_file, st.enabled INTO var_target_name_sns_arn_requires_file
                    FROM ubs_state.stage s
                    JOIN ubs_state.stage_topic st on s.id = st.stage_id
                    WHERE s.id = var_target.target_stage_id;

                    IF var_target_name_sns_arn_requires_file.enabled THEN
                         SELECT var_target_name_sns_arn_requires_file.name, var_target_name_sns_arn_requires_file.sns_arn, var_target_name_sns_arn_requires_file.requires_file
                                INTO var_next_stg;
                         var_next_stages = array_append(var_next_stages, var_next_stg);
                    end if;
                end if;

            END LOOP;

        -- the target is self (NOT_STARTED) --
        ELSIF in_stage_status = 'NOT_STARTED' THEN
            SELECT id into var_cur_stage_id
            FROM ubs_state.stage WHERE name = in_stage_name;

            SELECT ubs_state.check_stage_dependencies(var_main_id,
                                                     var_cur_stage_id)
                    INTO var_is_complete;

            IF var_is_complete THEN
                SELECT s.name, st.sns_arn, s.requires_file, st.enabled INTO var_target_name_sns_arn_requires_file
                FROM ubs_state.stage s
                JOIN ubs_state.stage_topic st on s.id = st.stage_id
                WHERE s.id = var_cur_stage_id;

                IF var_target_name_sns_arn_requires_file.enabled THEN
                    SELECT var_target_name_sns_arn_requires_file.name, var_target_name_sns_arn_requires_file.sns_arn, var_target_name_sns_arn_requires_file.requires_file
                                INTO var_next_stg;
                    var_next_stages = array_append(var_next_stages, var_next_stg);
                end if;
            end if;
        END IF;

        RETURN var_next_stages;
    END
$$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION update_stages_pipelines_get_targets(in_sha256 bytea,
                                                               in_stage_name varchar,
                                                               in_stage_status varchar,
                                                               in_stage_version_major smallint,
                                                               in_stage_version_minor smallint,
                                                               in_stage_completion_date timestamp with time zone)
RETURNS ubs_state.type_next_stage[]
AS $$
    DECLARE
        var_update_succeeded bool;
        var_next_stages ubs_state.type_next_stage[];

    BEGIN
        SELECT ubs_state.update_stage_pipeline_status(in_sha256,
                                             in_stage_name,
                                             in_stage_status,
                                             in_stage_version_major,
                                             in_stage_version_minor,
                                             in_stage_completion_date)
            INTO var_update_succeeded;

        IF NOT var_update_succeeded THEN
            RETURN var_next_stages;  -- return empty list
            -- TODO: return error message upon failure and bail before get next stage, controller needs to be able to handle errors
        end if;

        SELECT ubs_state.get_next_stages(in_sha256 ,
                               in_stage_name,
                               in_stage_status,
                               in_stage_version_major,
                               in_stage_version_minor)
            INTO var_next_stages;

        RETURN var_next_stages;
    END
$$ LANGUAGE plpgsql;

DO $$ BEGIN
    CREATE TYPE stage_summary AS (
        stage_name VARCHAR,
        stage_status VARCHAR,
        stage_version INT[]
    );
EXCEPTION
    WHEN duplicate_object THEN null;
END $$;

DO $$ BEGIN
    CREATE TYPE pipeline_summary AS (
        sha256 BYTEA,
        pipeline_name VARCHAR,
        pipeline_status VARCHAR,
        pipeline_version INT[],
        pipeline_stages stage_summary[]
    );
EXCEPTION
    WHEN duplicate_object THEN null;
END $$;

-- start pipeline or return current state. Restart if force_restart
CREATE OR REPLACE FUNCTION ubs_state.start_pipeline(in_sha256 bytea,
                                                    in_pipeline_name varchar,
                                                    force_restart bool
                                                    )
RETURNS ubs_state.pipeline_summary AS $$
    DECLARE
        in_sha256_main_id BIGINT;
        in_main_pipeline_link_id BIGINT;
        in_pipeline_status_id BIGINT;
        return_data ubs_state.pipeline_summary;
        not_started_status_id SMALLINT;
        pipeline_current_version_id SMALLINT;
        in_pipeline_id SMALLINT;
        var_stage_id SMALLINT;
    BEGIN
        -- we will need this constant
        SELECT id FROM ubs_state.status WHERE name = 'NOT_STARTED' INTO not_started_status_id;
        SELECT id FROM ubs_state.pipeline WHERE name = in_pipeline_name INTO in_pipeline_id;

        -- if requesting pipeline for file we've never seen, create an entry that it has been requested.
        INSERT INTO ubs_state.main (sha256) VALUES (in_sha256) ON CONFLICT (sha256) DO NOTHING;
        SELECT id FROM ubs_state.main WHERE sha256 = in_sha256 INTO in_sha256_main_id;

        -- create main_stage_links and NOT_STARTED statuses for those stages if missing rows.
        WITH missing_stage_links_for_pipeline AS (
            SELECT s.id stage_id, s.current_version_id version_id
            FROM ubs_state.pipeline_stage_link psl
            JOIN ubs_state.stage s on psl.pipeline_id = in_pipeline_id AND psl.stage_id = s.id
            LEFT JOIN ubs_state.main_stage_link msl on in_sha256_main_id = msl.main_id AND s.id = msl.stage_id
            WHERE msl.id IS NULL
        ),
        inserted_msls AS (
            INSERT INTO ubs_state.main_stage_link (main_id, stage_id)
            SELECT in_sha256_main_id, stage_id FROM missing_stage_links_for_pipeline
            ON CONFLICT (main_id, stage_id) DO NOTHING RETURNING id, stage_id
        )
            INSERT INTO ubs_state.stage_status (main_stage_link_id, version_id, status_id)
            SELECT im.id, mslp.version_id, not_started_status_id
            FROM inserted_msls im
            JOIN missing_stage_links_for_pipeline mslp ON mslp.stage_id = im.stage_id;

        -- create NOT_STARTED statuses for those stages that have not been run with their current version
        FOR var_stage_id IN (
            WITH outdated_stage_statuses AS (
                SELECT msl.id main_stage_link_id, s.id stage_id, s.current_version_id version_id
                FROM ubs_state.pipeline_stage_link psl
                JOIN ubs_state.stage s ON psl.pipeline_id = in_pipeline_id AND psl.stage_id = s.id
                JOIN ubs_state.main_stage_link msl ON s.id = msl.stage_id AND msl.main_id = in_sha256_main_id
                WHERE NOT exists(
                    SELECT 1 FROM ubs_state.stage_status ss
                    WHERE ss.main_stage_link_id=msl.id AND ss.version_id=s.current_version_id
                )
            ),
            inserted_stage_statuses  AS (
                INSERT INTO ubs_state.stage_status (main_stage_link_id, version_id, status_id)
                SELECT oss.main_stage_link_id, oss.version_id, not_started_status_id
                FROM outdated_stage_statuses oss
            )
                SELECT stage_id FROM outdated_stage_statuses
        )
        LOOP
            PERFORM ubs_state.update_ds_stage_status(var_stage_id, in_sha256_main_id);
        END LOOP;

        -- create pipeline if not exists
        INSERT INTO ubs_state.main_pipeline_link(main_id, pipeline_id) VALUES (in_sha256_main_id, in_pipeline_id)
        ON CONFLICT (main_id, pipeline_id) DO NOTHING;
        -- store mpl id for later.
        SELECT id FROM ubs_state.main_pipeline_link
        WHERE main_id = in_sha256_main_id
          AND pipeline_id = in_pipeline_id
        INTO in_main_pipeline_link_id;

        -- create pipeline status if not exists
        SELECT current_version_id FROM ubs_state.pipeline WHERE id = in_pipeline_id INTO pipeline_current_version_id;
        INSERT INTO ubs_state.pipeline_status(main_pipeline_link_id, version_id, status_id)
        VALUES (in_main_pipeline_link_id,
                pipeline_current_version_id,
                (SELECT id FROM ubs_state.status WHERE name = (
                    SELECT ubs_state.get_pipeline_status(in_pipeline_id, in_sha256_main_id)
                ))
                ) ON CONFLICT (main_pipeline_link_id, version_id) DO NOTHING;

        -- store pipeline status id for later
        SELECT id FROM ubs_state.pipeline_status
        WHERE main_pipeline_link_id = in_main_pipeline_link_id
          AND version_id = pipeline_current_version_id
        INTO in_pipeline_status_id;

        -- if restarting pipeline, stage statuses to not_started except pseudostages, then compute pipeline status.
        IF force_restart THEN
            UPDATE ubs_state.stage_status SET status_id = not_started_status_id WHERE id IN (
                SELECT ss.id FROM ubs_state.stage_status ss
                JOIN ubs_state.main_stage_link msl ON in_sha256_main_id = msl.main_id AND msl.id = ss.main_stage_link_id
                JOIN ubs_state.stage s ON msl.stage_id = s.id
                JOIN ubs_state.pipeline_stage_link psl ON in_pipeline_id = psl.pipeline_id AND s.id = psl.stage_id
                WHERE s.name NOT IN ('FILE_EXISTS', 'APC_UPLOAD_ALLOWED')
            );
            UPDATE ubs_state.pipeline_status
            SET status_id = (
                SELECT id FROM ubs_state.status WHERE name = (
                SELECT ubs_state.get_pipeline_status(in_pipeline_id, in_sha256_main_id)))
            WHERE id = in_pipeline_status_id;
        END IF;

        -- collect return data == pipeline version, status, stages, stage statuses, stage versions, etc
        WITH pipeline_info AS (
            SELECT p.name,
                   sts.name status_name,
                   ARRAY[v1.major, v1.minor] as pipeline_version,
                   p.id as pipeline_id
            FROM ubs_state.pipeline p
            JOIN ubs_state.main_pipeline_link mpl ON in_main_pipeline_link_id = mpl.id AND mpl.pipeline_id = p.id
            JOIN ubs_state.pipeline_status ps ON ps.main_pipeline_link_id = mpl.id AND ps.version_id = p.current_version_id
            JOIN ubs_state.version v1 ON ps.version_id = v1.id
            JOIN ubs_state.status sts ON ps.status_id = sts.id
            WHERE p.id = in_pipeline_id
            )
                SELECT INTO return_data
                    in_sha256 as sha256,
                    pipeline_info.name as pipeline_name,
                    pipeline_info.status_name as pipeline_status,
                    pipeline_info.pipeline_version as pipeline_version,
                    (SELECT ARRAY(
                        SELECT ROW(
                            s.name, sts.name, ARRAY[v2.major, v2.minor]
                        )::ubs_state.stage_summary
                        FROM pipeline_info pi
                        JOIN ubs_state.pipeline_stage_link psl ON pi.pipeline_id = psl.pipeline_id
                        JOIN ubs_state.stage s ON psl.stage_id = s.id
                        JOIN ubs_state.main_stage_link msl ON in_sha256_main_id = msl.main_id AND s.id = msl.stage_id
                        JOIN ubs_state.stage_status ss ON ss.main_stage_link_id = msl.id
                        JOIN ubs_state.version v2 ON v2.id = ss.version_id
                        JOIN ubs_state.status sts ON ss.status_id = sts.id
                        WHERE ss.version_id = s.current_version_id
                    )::ubs_state.stage_summary[]) as pipeline_stages
                    FROM pipeline_info;

        return return_data;
    END;
    $$ LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION ubs_state.get_pipeline_executable_root_stages(in_sha256 bytea,
                                                                         in_pipeline_name varchar
                                                                        )
RETURNS TABLE (stage_name varchar, sns_arn varchar, version smallint[]) AS $$
    DECLARE
        in_pipeline_id SMALLINT;
    BEGIN
        SELECT id FROM ubs_state.pipeline WHERE name = in_pipeline_name INTO in_pipeline_id;

        RETURN QUERY

        -- get "roots" == stages in the pipeline that are not a target of any other stage
        WITH root_stages AS (
            SELECT psl.stage_id
            FROM ubs_state.main m
            JOIN ubs_state.main_stage_link msl ON in_sha256 = m.sha256 AND m.id = msl.main_id
            JOIN ubs_state.pipeline_stage_link psl ON in_pipeline_id = psl.pipeline_id AND msl.stage_id = psl.stage_id
            LEFT JOIN ubs_state.stage_target st ON psl.stage_id = st.target_stage_id
            WHERE st.id IS NULL
        ),
        -- get all pseudostages in the pipeline that have been completed for that hash.
        completed_pseudostages AS (
            SELECT s.id
            FROM ubs_state.main m
            JOIN ubs_state.main_stage_link msl ON in_sha256 = m.sha256 AND m.id = msl.main_id
            JOIN ubs_state.stage_status ss ON msl.id = ss.main_stage_link_id
            JOIN ubs_state.stage s ON msl.stage_id = s.id
            JOIN ubs_state.pipeline_stage_link psl ON in_pipeline_id = psl.pipeline_id AND s.id = psl.stage_id
            JOIN ubs_state.status sts ON ss.status_id = sts.id
            WHERE sts.name = 'COMPLETE'
              AND s.name IN ('FILE_EXISTS', 'APC_UPLOAD_ALLOWED')
        ),
        -- get all stages that are in the pipeline and are dependent on a completed pseudostage
        pseudostage_targets AS (
            SELECT st.target_stage_id stage_id
            FROM ubs_state.stage_target st
            JOIN ubs_state.pipeline_stage_link psl ON in_pipeline_id = psl.pipeline_id AND psl.stage_id = st.target_stage_id
            JOIN completed_pseudostages cp ON cp.id = st.stage_id
        )
            -- with a union of non-pseudostage roots and stages in pipeline dependent on pseudostage roots,
            -- collect the stage names, SNS ARNs to run the stages, and current versions for the stages to run to start
            -- all the pipeline paths.
            SELECT s.name, topic.sns_arn, ARRAY[v.major, v.minor]
            FROM ubs_state.main m
            JOIN ubs_state.main_stage_link msl ON in_sha256 = m.sha256 AND m.id = msl.main_id
            JOIN ubs_state.stage_status ss ON msl.id = ss.main_stage_link_id
            JOIN (
                SELECT * FROM root_stages
                UNION
                SELECT * FROM pseudostage_targets
            ) target_stages ON target_stages.stage_id = msl.stage_id
            JOIN ubs_state.stage_topic topic ON topic.stage_id = target_stages.stage_id
            JOIN ubs_state.stage s ON topic.stage_id = s.id
            JOIN ubs_state.status sts ON ss.status_id = sts.id
            JOIN ubs_state.version v ON s.current_version_id = v.id
            WHERE topic.enabled IS TRUE AND sts.name IN ('NOT_STARTED', 'ERROR_RETRY') AND ss.version_id = s.current_version_id;
    END;
$$ LANGUAGE plpgsql;

/*
/*
    Retention policy information about how many latest records to keep in
    stage_status and batch size to delete
*/
create table IF NOT EXISTS ubs_state.data_retention_policy
(
    id                  bigserial
        primary key,
    batch_size          bigint                                 not null,
    retention_data_size bigint                                 not null,
    sproc_name          varchar,
    date_created        timestamp with time zone default now() not null,
    date_modified       timestamp with time zone default now() not null
);

alter table ubs_state.data_retention_policy
    owner to cdcubsroot;

-- This is for allowing to keep latest 7M records in stage_status and main_stage_link
INSERT INTO ubs_state.data_retention_policy(batch_size, retention_data_size, sproc_name) values (25000, 7000000, 'prune');

/*
    Holds history of data pruning
*/
create table IF NOT EXISTS  ubs_state.data_retention_stats
(
    id                   bigserial
        primary key,
    table_name           varchar,
    min_id               bigint                                 not null,
    max_id               bigint                                 not null,
    min_retention_id     bigint                                 not null,
    row_count            bigint                                 not null,
    deleted_count        bigint                                 not null,
    pruning_run_duration smallint                               not null,
    date_created         timestamp with time zone default now() not null
);

alter table ubs_state.data_retention_stats
    owner to cdcubsroot;



/*
    Prunes stage_status and main_stage_link tables based on data_retention_policy
*/
create or replace procedure ubs_state.prune()
    language plpgsql
as
$$
DECLARE
        var_retention_policy RECORD;
        var_stage_status_summary RECORD;
        var_main_stage_link_summary RECORD;
        var_affected_rows bigint = 0;
        var_start_time timestamp;
        var_end_time timestamp;
    BEGIN

        -- Read Data retention policy data
        SELECT  batch_size,  retention_data_size
        INTO  var_retention_policy
        FROM ubs_state.data_retention_policy  limit 1;

        -- Get stats from stage_status table
        SELECT  count(id) as row_count,
                min(id)  as min_Id,
                max(id)  as  max_Id,
                0 as min_retention_id,
                0 as deleted_row_count
        INTO var_stage_status_summary
        from ubs_state.stage_status;

        -- Get min id of retention data. The data with Ids less than this min_retention_id will be pruned or deleted
        SELECT min(Id) INTO  var_stage_status_summary.min_retention_id FROM
        (SELECT Id FROM ubs_state.stage_status order by Id desc limit var_retention_policy.retention_data_size) temp;

        SELECT now() INTO var_start_time;

        -- Prune stage status table (child table of main stage link)
        WHILE 1 = 1 LOOP

            DELETE
            FROM ubs_state.stage_status
            WHERE Id in (SELECT Id FROM ubs_state.stage_status
            WHERE Id <  var_stage_status_summary.min_retention_id LIMIT var_retention_policy.batch_size);

            GET DIAGNOSTICS var_affected_rows := ROW_COUNT;

            var_stage_status_summary.deleted_row_count = var_stage_status_summary.deleted_row_count + var_affected_rows;

            IF var_affected_rows = 0 THEN
                EXIT ;
            END IF;
        END LOOP;

        SELECT  now() INTO var_end_time;

        -- Update pruning stats
        INSERT INTO ubs_state.data_retention_stats(table_name,
                                                   min_Id,
                                                   max_Id,
                                                   min_retention_Id,
                                                   row_count,
                                                   deleted_count,
                                                   pruning_run_duration)
        VALUES ('ubs_state.stage_status',
                var_stage_status_summary.min_Id,
                var_stage_status_summary.max_Id,
                var_stage_status_summary.min_retention_id,
                var_stage_status_summary.row_count,
                var_stage_status_summary.deleted_row_count,
                Extract(epoch FROM (var_end_time - var_start_time))*1000);


        -- Get summary stats from main-stage-link
        SELECT  count(id) as row_count,
                min(id)  as min_Id,
                max(id)  as  max_Id,
                0 as min_retention_id,
                0 as deleted_row_count
        INTO var_main_stage_link_summary FROM ubs_state.main_stage_link;

        -- minimum id of the main_stage_link data that needs to be retained
        -- Anything smaller than this Id can be deleted
        SELECT coalesce(min(main_stage_link_id),0) INTO var_main_stage_link_summary.min_retention_id
        FROM ubs_state.stage_status;

        SELECT now() INTO var_start_time;
        -- Prune stage status table (child table of main stage link)
        WHILE 1 = 1 LOOP

            DELETE
            FROM ubs_state.main_stage_link
            WHERE Id in (SELECT Id FROM ubs_state.main_stage_link
            WHERE Id <  var_main_stage_link_summary.min_retention_id LIMIT var_retention_policy.batch_size);

            GET DIAGNOSTICS var_affected_rows := ROW_COUNT;
            var_main_stage_link_summary.deleted_row_count = var_main_stage_link_summary.deleted_row_count + var_affected_rows;

            IF var_affected_rows = 0 THEN
                EXIT ;
            end if;
        END LOOP;

        SELECT now() INTO var_end_time ;

        INSERT INTO ubs_state.data_retention_stats(table_name,
                                                   min_Id,
                                                   max_id,
                                                   min_retention_id,
                                                   row_count,
                                                   deleted_count,
                                                   pruning_run_duration)
        VALUES ('ubs_state.main_stage_link',
                var_main_stage_link_summary.min_Id,
                var_main_stage_link_summary.max_Id,
                var_main_stage_link_summary.min_retention_id,
                var_main_stage_link_summary.row_count,
                var_main_stage_link_summary.deleted_row_count,
                 Extract(epoch FROM (var_end_time - var_start_time))*1000 );

    END
$$;

alter procedure ubs_state.prune() owner to cdcubsroot;
*/
